-----------------------------------------
-- ALU
-----------------------------------------

---- Description --------------------------------------------------
-- ALU Function
---- Arithmetic
---- Compare
---- Logic
---- Shift
--
-- cf."߃Zbg߂܂傤"
-- OP is operation code (op2)
-- SRC1 := R2, SRC2 := R3 or Imm(sign extended);
-- R1 := DST;
-------------------------------------------------------------------

---- Notes --------------------------------------------------------
-- 2007.02.23
--   gݍ킹HŁB
--   ̓Iɂ́Awhen-elseɏBfunctionvSzcB
--------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_signed.all;

entity ALU is
    port( OP : in std_logic_vector(3 downto 0);          --- opcode
          SRC1,SRC2 : in std_logic_vector(31 downto 0);  --- R2 -> SRC1, R3 -> SRC2
          DST : out std_logic_vector(31 downto 0));      --- output
end ALU;


architecture RTL of ALU is
    
    --Shift߁BABbitshiftBOɂEVtgVtg𕪂ĂB
    function shift(A: std_logic_vector(31 downto 0);B: integer;O:std_logic_vector(1 downto 0)) return std_logic_vector is
   
      variable SINTL,SINTRL,SINTRA : Integer;
      constant ALL0 : std_logic_vector(31 downto 0) := (others => '0');
      constant ALL1 : std_logic_vector(31 downto 0) := (others => '1');
      variable RET : std_logic_vector(31 downto 0);
	 variable SUBR : std_logic_vector(63 downto 0);

      begin
        case O is
           when "00" => -- sl,sli
               if B >= 32 then
                   return ALL0;
               else
			    SUBR := A & ALL0;
			    return SUBR(31+B downto B);
               end if;
           when "01" => -- srl,srli
               if B >= 32 then
                   return ALL0;
               else
			    SUBR := ALL0 & A;
			    return SUBR(31+B downto B);
               end if;
           when "10" => -- sra,srai
               if B >= 32 then
                   RET := (others => A(31));
                   return RET;
               else
                   if A(31) = '1' then
				   SUBR := ALL1 & A;
				   return SUBR(31+B downto B);
                   else
				   SUBR := ALL0 & A;
				   return SUBR(31+B downto B);
                   end if;
               end if;
           when others => return "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
       end case;
 end shift;
 
 function mul(A:std_logic_vector(31 downto 0);B:std_logic_vector(31 downto 0);S:std_logic) return std_logic_vector is
    variable V : std_logic_vector(63 downto 0);
 begin
     V := A * B;
     if S = '0' then
         return V(31 downto 0);
     else
        return V(63 downto 32);
      end if;
 end mul;

function cmp(A:integer;B:integer;S:std_logic_vector(1 downto 0)) return std_logic_vector is
   variable ALL0 : std_logic_vector(31 downto 0) := (others => '0');
   variable ALL1 : std_logic_vector(31 downto 0) := (others => '1');
begin
    if S = "01" then
        if A = B then
            return ALL1;
        else
           return ALL0;
       end if;            
    elsif S= "10" then
       if A < B then
          return ALL1;
       else
          return ALL0;
       end if;        
    elsif S = "11" then
       if A > B then
           return ALL1;
       else
           return ALL0;
       end if;
    else
       return ALL0;
   end if;

end cmp;

 function div(A:integer;B:integer) return std_logic_vector is
    variable V,I : integer;
    variable BB : integer := B*2;
    variable BBB : integer := B*5;
    variable BBBB : integer := B*8;
 begin
 	if B = 2 then
	     V := A / 2;
		return CONV_STD_LOGIC_VECTOR(V,32);
	else
		if (A >= 0) and (A < B) then
			V := 0;
		elsif (A >= B) and (A < BB) then
			V := 1;
		elsif (A >= BB) and (A < B + BB) then
			V := 2;
		elsif (A >= BB+B) and (A < BB+BB) then
			V := 3;
		elsif (A >= BB+BB) and (A < BBB) then
			V := 4;
		elsif (A >= BBB) and (A < BBB+B) then
			V := 5;
		elsif (A >= BBB+B) and (A < BBB+BB) then
			V := 6;
		elsif (A >= BBB+BB) and (A < BBBB) then
			V := 7;
		elsif (A >= BBBB) and (A < BBBB+B) then
			V := 8;
		elsif (A >= BBBB+B) and (A < BBBB+BB) then
			V := 9;
		else
			V := 0;
		end if;

		return CONV_STD_LOGIC_VECTOR(V,32);
	end if;
 end div;

begin

    DST <= -- arith
            SRC1 + SRC2 when OP = "0000" else
            SRC1 - SRC2 when OP = "0001" else
            mul(SRC1,SRC2,'0') when OP = "0010" else
            --mul(SRC1,SRC2,'1') when OP = "0011" else
            div(CONV_INTEGER(SRC1),CONV_INTEGER(SRC2)) when OP = "0100" else
           -- compare
            cmp(CONV_INTEGER(SRC1),CONV_INTEGER(SRC2),OP(1 downto 0)) when OP(3 downto 2) = "01" else  -- eq,eqi
           -- logic
            SRC1 and SRC2 when OP = "1000" else -- and,andi
            SRC1 or SRC2 when OP = "1001" else -- or,ori
            SRC1 xor SRC2 when OP = "1010" else -- xor,xori
            not SRC1 when OP = "1011" else     -- not
            -- shift
            shift(SRC1,CONV_INTEGER('0' & SRC2(5 downto 0)),"00") when OP = "1100" else -- sl,sli
            shift(SRC1,CONV_INTEGER('0' & SRC2(5 downto 0)),"01") when OP = "1101" else -- srl,srli
            shift(SRC1,CONV_INTEGER('0' & SRC2(5 downto 0)),"10") when OP = "1110" else -- sra,srai
            (others => 'X');

end RTL;
